/*
   NAME
   freedup -- search for duplicate files in one or more directory hierarchies
   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2007 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 9 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.*/

/* GNU freedup was written by Andreas Neuper <ANeuper@web.de>  */
#include "sha1.h"
#include "freedup.h"

/* Name this program was called with. */
char *program_name;

void version(void)
{
    printf( "\n\tfreedup: Version " VERSION FLAVOUR " \2512007 by Andreas Neuper.\n"
		"\t sha1.c/h of sha-1.0.4 \2512001-2003 by Allan Saddi.\n\n");
}
static void printhashchoice(void);
static void printxtrachoice(void);
/* SYNOPSIS */
void usage(void)
{
    printf( "USAGE:\t%s [options] [<tree> ...]\n\n", program_name );
    printf( "\t-a\t   provide compatibility to freedups by William Stearns.[=-upg]\n"
   	    "\t-c\t   count file space savings per linked file.\n"
	    "\t-d\t   requires the modification time stamps to be equal.\n" 
	    "\t-f\t   requires the path-stripped file names to be equal.\n"
	    "\t-g\t   requires groups of the files to be equal before linking.\n" 
	    "\t-h\t   shows this help. [other option are ignored]\n"
	    "\t-i\t   Decide in interactive mode individually, what to do with them.\n"
	    "\t-l\t   only allow hardlinks. No symlinks are established.\n" 
	    "\t-m <bytes> only consider larger files. (if applicable use: -o \"-size +#\")\n" 
	    "\t-n\t   do not really perform links [no action].\n" );
    printf( "\t-o <opts>  pass an option string to the initially called find command.\n"
	    "\t-p\t   requires file permissions to be equal.\n" 
	    "\t-q\t   produces no output during the run (toggles -c and -v to off).\n" 
	    "\t-s\t   generate symlinks although some given paths are relative.\n"
	    "\t-t <type>  select hash method: ");
    printhashchoice();
    printf( "\t-u\t   requires users of the files to be equal before linking.\n" 
	    "\t-v\t   display shell commands to perform linking [verbose mode].\n"
	    "\t-V\t   display current version.\n"
	    "\t-w\t   only weak symbolic links allthough hardlinks might be possible.\n" 
	    "\t-x <style> select comparison style: " );
    printxtrachoice();
    printf( "\t-#\t   disable fast comparison using hash strings [default uses HASH].\n" 
	    "\t-0\t   disable linking of empty files i.e. files of size 0.\n" 
	    "\t<tree>\t   any directory to scan for duplicate files recursively.\n\n"
            "\tOptions are toggle switches. Their final state applies.\n"
	    "\tLater <tree> entries are linked to the earlier ones.\n"
	    "\tIf no <tree> is given, filenames are expected from standard input.\n"
	    "\tWhen standard input is used the option -o has no effect."
         );
    version();
    return;
}
   
/* OPTION DEFAULTS  */
	int showsave=0;	/* -c */
	int sametime=0;	/* -d */
	int samename=0;	/* -f */
	int samegrp =0;	/* -g */
	int hardlink=0;	/* -l */
	int minbytes=0;	/* -m */
	int testmode=0;	/* -n */
	int sameperm=0;	/* -p */
	int quietrun=0;	/* -q */
	int sameuser=0;	/* -u */
	int showmode=0;	/* -v */
	int weaklink=0;	/* -w */
	int interact=0; /* -i */
	int nonzero=0;	/* -0 */
	int usehash=1;	/* -# */
	int extratyp=0; /* -x */
	int nosymlinks=0;	/* -s */
	char*findopta=NULL;	/* -o */
	int hashmethod=0;	/* -t */

/* DESCRIPTION

   main() handles the command line parameters
	  scans for files to consider
	  builds a list of all files to consider
	  qsort()s the list of files
   comparison_by_size() is called for two files by qsort()
	  for equally sized files compares hash signature
	  comparison_by_content() when the same size & hash
   comparison_by_content() is only called by comparison_by_size()
	  reads and compares both files block by block
	  determines priority, i.e. later args count less
	  calls link_files() for identical files
   link_files() replaces the second file by a link to the first
	  determines whether both files are on the same fs
	  uses inode hard links on the same device number
	  uses symlinks if device numbers differ

   Additionally various helper functions are deployed.
*/

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
#include <assert.h>
#include <ctype.h>

void perror(const char *s);
#include <errno.h>
int errno;

/* internal counters */
/* link actions performed */
int totalinked=0;
int symlinked=0;
int deleted=0;
unsigned long long linksaved=0;

/* The number of regular files to scan. */
int frdinfosize=INITIALFINFOCOUNT;
int filecount=0;
int dupscount=0;

/* Time at start of execution.  */
time_t start_time;

/* Seconds between 00:00 1/1/70 and either one day before now
   (the default), or the start of today (if -daystart is given). */
time_t cur_day_start;

/* Status value to return to system. */
int exit_status;

/* The array pointer we use for sorting. */
frdinfo*file_info;

/* The array pointer we use for referencing duplicates in interactive mode. */
dinfo*dups_info;

#include "auto.h"

static void printxtrachoice(void)
{
    int n;
    unsigned char c=' ';
    for(n=1;extra[n].type!=NULL;n++)
    {
	if( extra[n].type != NULL )
	{
	    if(c==',') putchar(c); else c=',';
	    printf(" %s",extra[n].type);
	}
     /* else perror("lstat() failed while reading file statistics"); */
    }
    printf("\n");
    return;
}

int selectextra(const char*type)
{
    int method=0;
    assert(type!=0);
    while( extra[method].type != NULL )
    {
	if( strncmp(extra[method].type,type,strlen(type)) == 0 )
	{
	    break;
	} else method++;
    }
    if( extra[method].type == NULL )
    {
	if(showmode)
	    fprintf(stderr,
		"Cannot determine a extra comparison module called \"%s\".\n", type);
	return 0;
    }
    return method;
}

#ifdef HASHSUM
int hashreject=0;
int hashread=0;

struct hashtabelle {
	const char *type;
	const char *program;
	const int   length;
	const int   space1;
	const int   space2;
} hashme[] = {
     { "internal sha1","sha1",SHA1_HASH_SIZE, 40, 41 },
#if defined(SHA512SUM)
     { "sha512",       SHA512SUM,128,128,129 },
#endif
#if defined(SHA384SUM)
     { "sha384",       SHA384SUM, 96, 96, 97 },
#endif
#if defined(SHA256SUM)
     { "sha256",       SHA256SUM, 64, 64, 65 },
#endif
#if defined(SHA224SUM)
     { "sha224",       SHA224SUM, 56, 56, 57 },
#endif
     { "sha1",         SHA1SUM,   40, 40, 41 },
     { "md5",          MD5SUM,    32, 32, 33 },
/*   { "cksum",        CKSUM,     16, 16, 10 }, may not be used since
	variable length hash result is currently not supported */
     { "sum",          SUM,       11,  6,  8 }, /* not recommended */
     {  NULL,          NULL,       0,  0,  0 }
};

static void printhashchoice(void)
{
    int n;
    unsigned char c=' ';
    static struct stat xstat;
    for(n=1;hashme[n].type!=NULL;n++)
    {
	if( stat( hashme[n].program, &xstat) == 0)
	{
	    if( S_ISREG(xstat.st_mode) && ((S_IXOTH|S_IXGRP|S_IXUSR)&xstat.st_mode) )
	    {
		if(c==',') putchar(c); else c=',';
		printf("%s",hashme[n].type);
	    }
	}
     /* else perror("lstat() failed while reading file statistics"); */
    }
    printf("\n");
    return;
}

static void gethash_internal(frdinfo*a)
{
    char buffer[BLOCK+2];
    FILE*file2hash;
    SHA1CONTEXT context;
    size_t size,rdsz=0;

    hashread++;
    file2hash = fopen (a->name, "rb");
    if(file2hash == 0)
    {
	/* perror("fopen() failed - file was probably deleted"); */
	return;
    }
    SHA1INIT(&context);
    fseek( file2hash, extra[extratyp].start(((frdinfo*)a)->name), SEEK_SET );
    rdsz = fread(buffer,1,(a->size<BLOCK)?a->size:BLOCK,file2hash);
    size = a->size - rdsz;
    do {
        if( ferror(file2hash) )
        {
	    fclose(file2hash);
	    return;
        }
        SHA1UPDATE(&context, buffer, rdsz);
        rdsz = fread(buffer, 1, (size<BLOCK)?size:BLOCK, file2hash);
        size = size - rdsz;
    } while( (!feof(file2hash))&&(rdsz!=0) );
    fclose(file2hash);
    a->hash = calloc(SHA1_HASH_SIZE+2,sizeof(char));
    SHA1FINAL(&context, (unsigned char*)a->hash);
    return;
}

static void gethash(frdinfo*a)
{
    char buffer[MAXPATHLENGTH+132];
    FILE*file2hash;
    int n=0;
    hashread++;
    /*
     * this routine uses external shell commands
     */
    if( hashmethod==0 || extratyp!=0 )
    {
	gethash_internal(a);
	return;
    }
    sprintf(buffer,"%s \"%s\" 2>/dev/null",hashme[hashmethod].program, a->name);
    file2hash = popen (buffer, "r");
    if(file2hash == 0)
    {
	perror(buffer);
	exit(-1);
    }
    fgets(buffer,MAXPATHLENGTH+132,file2hash);
    a->hash = (char*)calloc(hashme[hashmethod].length+2,sizeof(char));
    if( feof(file2hash) )
    {
	/*  hash was unsuccessful;
	 *  usually fails due to special characters in the file name
	 *  (this was at least the reason for redirecting errors)
	 *  write std dummy value;
	 */
        for(n=0;n<hashme[hashmethod].length;n++)
	   (a->hash)[n]='0';
	/*  it would be better (correct)
	 *  to link it to some temporary file,
	 *  to evaluate it, and to continue then
	 */
    } else
    {
	/*  if everything works fine
	 *  there should appear 32 characters
	 *  followed by two spaces and the file name
	 */
        assert(buffer[hashme[hashmethod].space1]==' ');
        assert(buffer[hashme[hashmethod].space2]==HASHCHAR);
        if(0) printf("HASH=%s\n", buffer);
	/* store the hash value to avoid recalculation */
        strncpy(a->hash,buffer,hashme[hashmethod].length);
    }
    pclose(file2hash);
    /* the string was already initialised with zero
     * since I am somewhat paranoid ...
     */
    a->hash[hashme[hashmethod].length]=0;
    return;
}

int hashprobe(const int method,const char*filename)
{
    static struct stat astat;
    char buffer[MAXPATHLENGTH+132];
    FILE*file2hash;

    if(stat(hashme[method].program, &astat) != 0)
    {
	/* the program does not even exist */
	sprintf(buffer,"%s: No hash program at %s",hashme[method].type,
		hashme[method].program);
	perror(buffer);
	return(-1);
    } else
    {
	/* hash program should be a/point to regular file */
	if( ! S_ISREG(astat.st_mode) ) return -2;
	/* a program should be executable by someone */
	if( ! ((astat.st_mode)&(S_IXUSR|S_IXGRP|S_IXOTH)) ) return -2;
    }
    /*
     * this routine uses external shell commands
     */
    sprintf(buffer,"%s \"%s\" 2>/dev/null",
	hashme[method].program, filename);
    file2hash = popen (buffer, "r");
    if(file2hash == 0)
    {
	perror(buffer);
	return(-3);
    }
    fgets(buffer,MAXPATHLENGTH+132,file2hash);
    pclose(file2hash);
    if( strlen(buffer) < hashme[method].length )
	return -4;
    if( buffer[hashme[method].space1] != ' ' 
    ||  buffer[hashme[method].space2] != HASHCHAR )
    {
	if(showmode)
	    fprintf(stderr,
		"hash result is: \"%s\"\n"
		"%s: expected output format does not match template."
		" \"<#%d>%c%c<file>\"\n", buffer, hashme[method].type,
		hashme[method].length,hashme[method].space1,
		hashme[method].space2);
	return -5;
    }
    return 0;
}

int selecthash(const char*type)
{
    hashmethod=0;
    while(hashme[hashmethod].program!=NULL)
    {
	if( strncmp(hashme[hashmethod].type,type,strlen(type)) == 0)
	{
	    break;
	} else hashmethod++;
    }
    if( hashme[hashmethod].program==NULL)
    {
	if(showmode)
	    fprintf(stderr,
		"Cannot determine a hash function called \"%s\".\n", type);
	hashmethod=-1;
	return hashmethod;
    }
    if( hashprobe(hashmethod,program_name) == 0)
    {
	return hashmethod;
    } else
    {
	if(showmode)
	    fprintf(stderr,
		"Hash function \"%s\" oputput format not supported.\n",
		hashme[hashmethod].type);
	hashmethod=-1;
    }
    return hashmethod;
}

#endif /* HASHSUM */

void ttydirect(int on)
{
    static TTY new_io,old_io;
    int stty=(ECHO|ICANON);

    if(on==1)
    {
	if((tcgetattr(STDIN_FILENO, &old_io)) == EXIT_SUCCESS)
	{
	    new_io = old_io;
	    new_io.c_lflag = new_io.c_lflag & ~stty;
	    new_io.c_cc[VMIN] = 1; /* only one byte */
	    new_io.c_cc[VTIME]= 0; /*  immediately  */
	    tcsetattr(STDIN_FILENO, TCSANOW, &new_io);
	} else
	{
	    fprintf(stderr,"Could not switch to raw mode!\n");
	}
    } else
    {
	tcsetattr(STDIN_FILENO, TCSANOW, &old_io);
    }
}

int statcpy(frdinfo*a)
{
    static struct stat*astat;
    astat=&(a->info);
    if(lstat(a->name, astat) != 0)
    {
	perror("lstat() failed while reading file statistics");
	return(-1);
    }
    a->size = (extratyp==0) ? a->info.st_size : extra[extratyp].size(a->name);
    if(0)
    fprintf(stderr,">>%s>> %d - %d - %d - %d - %d\n",a->name,(int)(a->size) ,(int)(a->info.st_size),
	(int)(extra[extratyp].start(a->name)),(int)(extra[extratyp].end(a->name)),
	(int)(extra[extratyp].size(a->name)) );
    return 0;
}

char*symname(char*a,char*b)
{
    static char*name=NULL, *ptr;
    int n,hierarchies=0;
    
    if(a[0]=='/') return a;		  /* absolute Pathnames will be kept */
    for( n=0; a[n]==b[n] && a[n]!=0; n++); /* here Filenames start to differ */
    if( a[n]!=b[n] ) n--; 	/* avoid difficult situations with directory */
    for(;a[n]!='/' && n>0;n--);		      /* back to last directory mark */
    if(n!=0 && a[n]=='/')n++;			     /* should never be true */
    ptr=&b[n];			    /* upto here the pathnames are identical */
    for(hierarchies=0;ptr!=NULL;hierarchies++)
    {		/* determine how many dirlevels the target is located deeper */
	if( *ptr=='/' && ptr[1]=='/' )
	    hierarchies--;			  /* ignore multiple slashes */
	ptr=strchr(ptr,'/');
	if(ptr!=NULL && *ptr=='/')
	    ptr=ptr+1;				     /* avoids endless loops */
    }
    ptr=&a[n];	    /* symlink must point the determined hierarchy levels up */
    if(hierarchies>1)
    {
	if(name!=NULL) free(name);
	name=calloc(strlen(ptr)+3*hierarchies,sizeof(char));
	for(;hierarchies>1;hierarchies--) strcat(name,"../");
	ptr=strcat(name,ptr);			  /* before aiming to target */
    }
    return ptr;
}

void linkfilesecure(char*aname,char*bname,int(*lnk)(const char *, const char *))
{
    char tmpfilename[MAXPATHLENGTH+16],*symaname=aname;
    static struct stat tstat;
    /* 
     * generate unique tmp file name
     * move file to tmp ;
     * later link it ;
     * finally remove tmp
     */
    if( !testmode )
    {
	strcpy(tmpfilename,bname);
	strcat(tmpfilename,".XXXXXX");
	/*
	 * Please note that the warning on mktemp() is acceptable. 
	 * Since we intend to move the file we cannot create it with mkstemp().
	 */
	if( mktemp(tmpfilename) == NULL )
	{
	    perror("There is no unique temporory file name.");
	}
	if( lstat(tmpfilename,&tstat) != 0 )
	{
	    /*
	     * The errror needs not to be catched, since it is wanted
	     * that no file exists with the target name
	     */
	    rename( bname, tmpfilename );
	} else
	{
	    fprintf(stderr,"Not able to rename \"%s\" to \"%s\".",bname,tmpfilename);
	}
	if( lnk == symlink )
	{
	    symlinked++;
	    symaname=symname( aname, bname );
	}
	if( lnk( symaname, bname ) != 0 )
	{
	    perror("Linking failed.");
	}
	if( unlink( tmpfilename ) != 0 )
	{
	    perror("Unlink failed.");
	}
    }
    if( showmode || interact)
    {
	printf("ln %s \"%s\" \"%s\"\n",( lnk == symlink )?"-s":"-f",symaname,bname);
    }
    return;
}

int interaction(frdinfo*a, frdinfo*b)
{
    int n=0,m=0;

    /* merge or create dups_info entries */
    if( (a->ident == NULL) && (b->ident == NULL) )
    {
	dups_info[dupscount].name = (char**)calloc(32,sizeof(char**));
	dups_info[dupscount].name[0] = a->name;
	dups_info[dupscount].name[1] = b->name;
	/* other pointers are NULL, since calloc sets to zero */
	a->ident = b->ident = &dups_info[dupscount];
	dupscount++;
    } else
    {
	if( a->ident == NULL )
	{
	    /*clean up tangling references */
	    if( (b->ident->replace!=NULL) )
	    {
		b->ident = b->ident->replace;
	    }
	    /* simply inherit reference list and append the new filename */
	    a->ident = b->ident;
	    for(n=0;b->ident->name[n]!=NULL && n<32;n++);
	    b->ident->name[n] = a->name;
	} else
	if( b->ident == NULL )
	{
	    /*clean up tangling references */
	    if( (a->ident->replace!=NULL) )
	    {
		a->ident = a->ident->replace;
	    }
	    /* simply inherit reference list and append the new filename */
	    b->ident = a->ident;
	    for(n=0; a->ident->name[n]!=NULL && n<32;n++);
	    a->ident->name[n] = b->name;
	} else
	{
	    /*clean up tangling references */
	    while( ((a->ident->replace)!=NULL) )
	    {
		a->ident = a->ident->replace;
	    }
	    while( ((b->ident->replace)!=NULL) )
	    {
		b->ident = b->ident->replace;
	    }
	    /* merge the reference lists */
	    assert( a->ident->name!=NULL );
	    assert( b->ident->name!=NULL );
	    if( a->ident != b->ident )
	    {
		for(n=0;a->ident->name[n]!=NULL && n< 32   ;n++);
		for(m=0;b->ident->name[m]!=NULL && m<(32-n);m++) 
			a->ident->name[n+m] = b->ident->name[m];
		assert((n+m)<32);
		free(b->ident->name);
		b->ident->name = NULL;
		/* avoid active search for tangling references */
		b->ident->replace = a->ident;
		b->ident = a->ident;
	    }
	}
    }
    return 0;
}

int dupinfo_entry(dinfo*a)
{
    static struct stat astat;
    static int n,m,min=0,max=0,cnt=0,source,key;
    for( n=0; a->name[n]!=NULL; n++ )
    {
	if( lstat(a->name[n], &astat) != 0 )
	{
	    /* ignore this file, probably deleted */
	    continue;
	}
	if( testmode )
	{
	    printf( "%s\n", a->name[n] );
	}
	else
	{
	    printf( " %d : %c [%04x:%06x:%04o:%01d] %s\n",
		n, n+'A',
		(int)astat.st_dev,
		(int)astat.st_ino,
		(int)S_PERM(astat.st_mode),
		(int)astat.st_nlink,
		a->name[n]);
	}
    }
    if( !testmode )
    {
        static char buffer[MAXPATHLENGTH],target[32];
	/* 
	 * analyse and execute the answer
	 */
	printf("Delete on number, Link on letter, Symlink on Capital (first is source)\t");
	strncpy(target,EMPTYSTRING,n);
	source=-1;
	target[n]=0;
	buffer[0]=buffer[1]=buffer[2]=buffer[3]=0;
	ttydirect(1);
	while(buffer[0]!='\n') {
	    fflush(stdout);
	    //fgets(buffer, MAXPATHLENGTH, stdin);
	    buffer[0]=getchar();
	    for(m=0;buffer[m]!=0 && m<MAXPATHLENGTH;m++)
	    {
		key=buffer[m];
		if( isalpha(key) && source==-1 )
		{   source=toupper(key)-'A';
		    if(source>=n) source=-1;
		} else {
		if( isupper(key) ) { max++; if(key-'A'<n) target[(key-'A')]='S'; }
		if( islower(key) ) { min++; if(key-'a'<n) target[(key-'a')]='L'; } }
		if( isdigit(key) ) { cnt++; if(key-'0'<n) target[(key-'0')]='D'; }
		if( key=='\033'  ) { strncpy(target,EMPTYSTRING,n); source=-1;}
	    }
	    printf(EMPTYSTRING "\r  $ "); fflush(stdout);
	    printf(EMPTYSTRING "\r  $ ");
	    if( source!=-1 )
	    {
		if(!strncmp(EMPTYSTRING,target,n))
		    printf("\r <RETURN> to confirm. <ESC> to clear. All links will point to <%c>.",source+'a');
		target[source] = 'Q';
	    }
	    for(m=0;m<n;m++)
	    {
		switch(target[m])
		{
		    case 'D': printf("rm <%d>; ", m); break;
		    case 'S': printf("ln -s <%c> <%c>; ",source+'A', m+'A'); break;
		    case 'L': printf("ln <%c> <%c>; ",source+'a', m+'a'); break;
		}
	    }
	}
	ttydirect(0);
	if( (source==-1) && (max+min>0) ) return 1; /* all other cases are treated as valid */
        if( (max>0) && ((a->name[source])[0] != '/') )
	{
	    if(0) printf("DANGER: Resulting symlinks can be invalid due to relative pathname!");
	}
	/* 
	 * tell user what we will do
	 */
	puts("");
	for( n=0; a->name[n]!=NULL; n++ )
	{
	    if( lstat(a->name[n], &astat) != 0 )
	    {
		/* ignore this file, probably deleted */
		continue;
	    }
	    switch(target[n])
	    {
		case 'S':  printf("\t$ ");
			   linkfilesecure(a->name[source],a->name[n],symlink);
			   break;
		case 'L':  printf("\t$ ");
			   linkfilesecure(a->name[source],a->name[n],link);
			   break;
		case 'D':  printf( "\t$ rm -f %s\n", a->name[n] );
			   unlink( a->name[n] );
			   break;
		case 'Q':  break;
		case ' ':  printf( "\t$ #keep %s\n", a->name[n] );
			   break;
		default :  break;
	    }
	}
    }
    return 0;
}

void print_dupinfo(void)
{
    int n=0;
    for(n=0;n<dupscount;n++)
    {
	if(dups_info[n].name!=NULL)
	{
	    if(!testmode)
		printf("Del:Lnk [dev:i-node:perm:L] <filename>\n");
		if(dupinfo_entry(&dups_info[n])!=0) n--;
		printf("\n");
	}
    }
}

void 
link_files(frdinfo*a, frdinfo*b)
{
#ifdef DEBUG
    fprintf(stderr,"testmode=%d,showmode=%d,weaklink=%d,hardlink=%d,nosymlinks=%d,showsave=%d\n",
	testmode,showmode,weaklink,hardlink,nosymlinks,showsave);
#endif /*DEBUG*/
    if( S_ISLNK(a->info.st_mode) || S_ISLNK(b->info.st_mode) )
    {
        /* either one is already symlinked
	 * this avoids building up symlink chains (in case of a)
	 * or alternating symlinks during sort (in case of b)
	 */
	return;
    }
    if(nosymlinks && (a->info.st_dev != b->info.st_dev) )
    {
        /* when on different devices and symlinks are not allowed
	 * do not link, i.e. no also not prepare to link.
	 * This needs to be catched before any execution.
	 */
	return;
    }
    if(nonzero && (a->info.st_size == 0) )
    {
        /* when only non-empty files are allowed to be linked
	 * do not link, when they are identical and has size zero.
	 * This needs to be catched before any execution.
	 */
	return;
    }
    if( interact )
    {  /*
        * interaction() just prepares for interaction
	* i.e. accumulates information, but not do anything
	*/
	interaction(a,b);
    } else
    {
	if( !weaklink && ( a->info.st_dev == b->info.st_dev ) )
	{
	    /* 
	     * either st_dev or st_ino are different,
	     * was checked earlier prior to call
	     */
	    linkfilesecure(a->name,b->name,link);
	    statcpy(b);
	} else
	{
	    assert( weaklink || !nosymlinks );
	    if(!hardlink )
	    {
		linkfilesecure(a->name,b->name,symlink);
		statcpy(b);
	    }
	}
	if( showsave ) 
	{
	    printf("Linking \"%s\" saves %d bytes\n",
		    basename(b->name), 
		    (int)(b->info).st_size);
	}
	linksaved+=(int)(b->info).st_size;
	totalinked++;
    }
    return;
}

int comparison_by_content(const void*a,const  void*b)
{
    FILE*afile=NULL,
	*bfile=NULL;
    char ablock[BLOCK+2],
	 bblock[BLOCK+2];
    int  aread,
	 bread,
	 result;
    size_t asize, bsize;
    assert(a!=NULL);
    assert(b!=NULL);
    afile=fopen(((frdinfo*)a)->name,"r");
    if( afile == NULL )
    {
	/* perror("fopen() failed - file was probably deleted"); */
	return(-1);
    }
    bfile=fopen(((frdinfo*)b)->name,"r");
    if( bfile == NULL )
    {
	/* perror("fopen() failed - file was probably deleted"); */
	return(1);
    }
    fseek(afile,extra[extratyp].start(((frdinfo*)a)->name),SEEK_SET);
    fseek(bfile,extra[extratyp].start(((frdinfo*)b)->name),SEEK_SET);
    asize=(((frdinfo*)a)->size);
    bsize=(((frdinfo*)b)->size);
    while(!(feof(afile)||feof(bfile)||asize==0||bsize==0))
    {
        aread = fread(ablock,1,(asize<BLOCK)?asize:BLOCK,afile);
        asize = asize - aread;
	assert((aread!=0)||feof(afile)||(asize==0));
        bread = fread(bblock,1,(bsize<BLOCK)?bsize:BLOCK,bfile);
        bsize = bsize - bread;
	assert((bread!=0)||feof(bfile)||(bsize==0));
	assert(aread==bread||feof(afile)||feof(bfile));
	result=memcmp(ablock,bblock,aread);
	if( result!=0 )
	{
	    fclose(afile);
	    fclose(bfile);
	    return(result);
        }
    }
    fclose(afile);
    fclose(bfile);
    /* decide which has higher prio by
     * 1) collect links where more hardlinks exist
     * 2) link to the one earlier named in command line
     * 3) arbitrary /# inactive: shorter path name #/
     * NOT: shorter path length /+/ may result in alternating linkage
     * the one with lower priority will be deleted
     */
    if(((frdinfo*)a)->info.st_nlink > ((frdinfo*)b)->info.st_nlink )
    {
	/* We should elimate unlinked files and
	 * keep more often linked files as collectors
	 */
	link_files(((frdinfo*)a),((frdinfo*)b));
    } else
    {
	if(((frdinfo*)a)->info.st_nlink < ((frdinfo*)b)->info.st_nlink )
	{
	    link_files(((frdinfo*)b),((frdinfo*)a));
	} else
	{
	    if(((frdinfo*)a)->prio<((frdinfo*)b)->prio)
	    {
		link_files(((frdinfo*)a),((frdinfo*)b));
	    } else
	    {
		if(((frdinfo*)a)->prio>((frdinfo*)b)->prio)
		{
		    link_files(((frdinfo*)b),((frdinfo*)a));
		} else
		{
/**/
/* it did not work well to give priority within one file system tree	*/
/*	    if(strlen(((frdinfo*)a)->name)<strlen(((frdinfo*)b)->name))	*/
/**/
		    link_files(((frdinfo*)a),((frdinfo*)b));
		}
	    }
	}
    }
    return(0);
} 

static int comparison_by_size(const void*a, const void*b)
{
    static int sizediff;
    assert(a!=NULL);
    assert(b!=NULL);
    sizediff = (( (((frdinfo*)a)->size) 
                - (((frdinfo*)b)->size) ));
    /*
     * avoid work when not desired
     * follow the minbytes directive
     * "-m" does not work on payload
     */
    if((minbytes > (((frdinfo*)a)->info.st_size))
    || (minbytes > (((frdinfo*)b)->info.st_size)))
      {
        return(sizediff);
      }
    /*
     * up to here it is obvious:
     * - different size different file
     * - same size -> differs file?
     */
    if(sizediff == 0)
    {
	/* exclude weaklinks from linking, must not result in endless link loops */
	if( (((frdinfo*)a)->info.st_mode & S_IFMT) != S_IFREG 
	 || (((frdinfo*)b)->info.st_mode & S_IFMT) != S_IFREG )
	    /* same size but no linking allowed */
	    return 0;
	/* compare the path-less file names
	 * if the samename directive (-f) requires this */
	if( strcmp(basename(((frdinfo*)a)->name),basename(((frdinfo*)b)->name)) 
	 && (samename==1) )
	    /* different names do not allow linking */
	    return strlen(basename(((frdinfo*)a)->name))-strlen(basename(((frdinfo*)b)->name));
	/* check whether the user entries match (-u) */
	if( ((((frdinfo*)a)->info.st_uid)!=(((frdinfo*)b)->info.st_uid)) 
	 &&  (sameuser==1) )
	    /* different user ids do not allow linking */
	    return ((((frdinfo*)a)->info.st_uid)-(((frdinfo*)b)->info.st_uid));
	/* check whether the group entries match (-g) */
	if( ((((frdinfo*)a)->info.st_gid)!=(((frdinfo*)b)->info.st_gid)) 
	 &&  (samegrp ==1) )
	    /* different group ids do not allow linking */
	    return ((((frdinfo*)a)->info.st_gid)-(((frdinfo*)b)->info.st_gid));
	/* check whether the permissions match (-p) */
	if( (((((frdinfo*)a)->info.st_mode)&(ALLPERMS))!=((((frdinfo*)b)->info.st_mode)&(ALLPERMS))) 
	 && (sameperm==1) )
	    /* different permissions do not allow linking */
	    return 0;
	/* check whether the modification dates match (-d) */
	if( ((((frdinfo*)a)->info.st_mtime)!=(((frdinfo*)b)->info.st_mtime)) 
	 && (sametime==1) )
	    /* different modification time do not allow linking */
	    return((((frdinfo*)a)->info.st_mtime)-(((frdinfo*)b)->info.st_mtime)); 
	/* check whether already hardlinked */
	if( ((((frdinfo*)a)->info.st_ino != ((frdinfo*)b)->info.st_ino))
	 || ((((frdinfo*)a)->info.st_dev != ((frdinfo*)b)->info.st_dev)) )
	{
#ifdef HASHSUM
	    if( usehash )
	    {
		if((((frdinfo*)a)->hash)==NULL) gethash((frdinfo*)a);
		if((((frdinfo*)b)->hash)==NULL) gethash((frdinfo*)b);
		assert(((frdinfo*)a)->hash != NULL);
		assert(((frdinfo*)b)->hash != NULL);
		if( strncmp( (((frdinfo*)a)->hash), (((frdinfo*)b)->hash), hashme[hashmethod].length )!=0 )
		{
		    hashreject++;
		    return( strncmp( (((frdinfo*)a)->hash), (((frdinfo*)b)->hash), hashme[hashmethod].length ) );
		} else
		{
		    comparison_by_content(a,b);
		}
	    } else
#endif
	    {
		comparison_by_content(a,b);
	    }
	}
    }
    return(sizediff);
}

void print_infolist(void)
{
    int n;
    for(n=0;n<filecount;n++)
    {
        printf("%9d %x:%05d %s\n",
	  (int)file_info[n].info.st_size,
	  (int)file_info[n].info.st_dev,
	  (int)file_info[n].info.st_ino,
	  file_info[n].name);
    }
}

void print_duplist(void)
{
    int n,m,identicalcount=0,alreadylinked=0;
    for(n=0;n<filecount;n++)
    {
        for(m=1;n+m<filecount && (file_info[n].size==file_info[n+m].size);m++)
        {
	    identicalcount=m+1;
	    if( (file_info[n].info.st_dev  == file_info[n+m].info.st_dev)
	    &&  (file_info[n].info.st_ino  == file_info[n+m].info.st_ino) )
	    {
		alreadylinked+=1;
	    }
if(0)	    printf("%01d\t%01d\t",identicalcount,alreadylinked);
	}
	if( (identicalcount==m) && (identicalcount==alreadylinked+1) )
	{
	    identicalcount=0;
	} 
 	alreadylinked=0;
	{
	printf("%9d %x:%05d %s\n",
	  (int)file_info[n].info.st_size,
	  (int)file_info[n].info.st_dev,
	  (int)file_info[n].info.st_ino,
	  file_info[n].name);
	    identicalcount--;
	}
    }
}

void printstats(void)
{
    if(deleted) fprintf(stderr,"%d Files have been%s deleted\n",deleted,testmode?" marked to be":"");
    fprintf(stderr,"%d file%s of %d %sreplaced by links.\nThe total size of replac%s files %s %lld bytes.\n",
	totalinked,(totalinked==1)?"":"s",filecount,testmode?"will be ":"",
	testmode?"eable":"ed", testmode?"is":"was", linksaved);
#ifdef HASHSUM
    if( usehash )
    fprintf(stderr,"%s hash algorithm had to read %d files to avoid %d file comparisons.\n",
	hashme[hashmethod].type,hashread,hashreject);
#endif /* HASHSUM */
}


int readtree( frdinfo**fin, int*total, const int filecount, FILE*in)
/*
 * "total" will be modified
 * "fin" might be modified
 * "in" is read until EOF
 */
{
    char buffer[MAXPATHLENGTH];
    int  count;

    for(count=filecount;!feof(in);count++)
    {
	fgets(buffer,MAXPATHLENGTH,in);
	if( !feof(in) && (strlen(buffer)>1) )
	{
	    (*fin)[count].prio=count;
	    (*fin)[count].name=(char*)malloc(strlen(buffer)+2);
	    if( (*fin)[count].name == NULL )
	    {
		perror("malloc() for buffer in file info structure");
		exit(-1);
	    }
	    assert(buffer[strlen(buffer)-1]=='\n');
	    /* fails if MAXPATHLENGTH is exceeded */
	    buffer[strlen(buffer)-1]=0;
	    strcpy((*fin)[count].name,buffer);
	    (*fin)[count].ident=NULL;
	    if( statcpy(&((*fin)[count]))!=0
	    /*
	     * Drop symbolic links and devices here
	     */
	    ||  !S_ISREG((*fin)[count].info.st_mode) )
	    {
		free((*fin)[count].name);
		count--;
		continue;
	    }
	    if( count > (*total)-2 )
	    {   /*
		 * expand file_info array to take more members
		 */
		(*total) = (*total) + INITIALFINFOCOUNT;
		(*fin) = realloc((*fin), sizeof(frdinfo) * ( (*total) +4 ) );
		if( (*fin) == NULL )
		{
		    perror("realloc() to extend file info structure");
		    exit(-1);
		}
	    }
	} else count--;
    }
    return count;
}

int
main (int argc, char *argv[])
{
  int c,i,lastlinked=0,remember=0;
  FILE*liste;
  char buffer[MAXPATHLENGTH],stopornot[128];
  program_name = argv[0];

#ifdef HAVE_SETLOCALE
  setlocale (LC_ALL, "");
#endif
#ifdef DEBUG
  /* Locale is not used in current implementation
   * Time is not used in current implementation
   */
  start_time = time (NULL);
  if(!quietrun) printf ("cur_day_start = %s", ctime (&cur_day_start));
#endif /* DEBUG */
  exit_status = 0;

    while ((c = getopt(argc, argv, "acdfghilm:no:pqsut:vVwx:#0")) != EOF) 
    {
        switch (c)
	{
	  case 'a': sameperm   = sameuser = samegrp = 1; break;
	  case 'c': showsave   = 1-showsave;   break;
	  case 'd': sametime   = 1-sametime;   break;
	  case 'f': samename   = 1-samename;   break;
	  case 'g': samegrp    = 1-samegrp ;   break;
	  case 'h': usage();                 exit(0);
	  case 'i': interact   = 1-interact;   break;
	  case 'l': hardlink   = 1-hardlink;   break;
	  case 'm': minbytes   = atoi(optarg); break;
	  case 'n': testmode   = 1-testmode;   break;
	  case 'o': findopta   = calloc( strlen(optarg)+2, sizeof(char) );
		    strcpy( findopta, optarg );break;
	  case 'p': sameperm   = 1-sameperm;   break;
	  case 'q': quietrun   = 1-quietrun;   showsave = showmode = 0; break;
	  case 's': nosymlinks = 1-nosymlinks; break;
	  case 't': hashmethod = selecthash(optarg);
					       break;
	  case 'u': sameuser   = 1-sameuser;   break;
	  case 'v': showmode   = 1-showmode;   break;
	  case 'V': version();                 exit(0);
	  case 'w': weaklink   = 1-weaklink;   break;
	  case 'x': extratyp   = selectextra(optarg);
					       break;
	  case '#': usehash    = 1-usehash;    break;
	  case '0': nonzero    = 1-nonzero;    break;
	  default:
		fprintf(stderr,"wrong option: \"%c\"\n",c);
		usage();
		exit(-1);
	}
    }
  
    if(usehash && hashmethod!=0)
    {
	if( hashme[hashmethod].program != NULL )
	{
	    if( hashprobe(hashmethod,program_name) != 0 )
	    {
		hashmethod++;
	    } /* else break;*/
	}
	if( hashme[hashmethod].program == NULL )
	{
	    usehash=0;
	    fprintf(stderr,"No working hashmethod found.\n--> Use of hash methods is disabled.\n\n");
	}
    }
    if(!testmode && extratyp!=0) {
	fprintf(stderr,"Extra Style comparisons are not thouroughly tested yet.\n"
	"You may loose header information, if you press <return>.\nUse CTRL-C to stop.");
	scanf("%s",stopornot); }
  /* 
   * Collect files to scan.
   * Read them and fill in stat now
   */
  file_info = (frdinfo*)calloc(frdinfosize+4,sizeof(frdinfo));
  if( file_info == NULL )
  {
      perror("calloc() to provide file info structure");
      return(-1);
  }
  if( argc == optind )
    {	/*
	 * Read file names (not dir names) from stdin
	 */
	if(!quietrun) fprintf(stderr,"Taking file names from stdin\n\n");
	filecount = readtree( &file_info, &frdinfosize, 0, stdin );
	/*
	 * Now we will skip the following for-loop
	 */
    } else
    {
      for (i=optind,filecount=0;i < argc;i++)
	{
	  if( argv[i][0] != '/' && (argc-optind) > 1 )
	    {
	      if(!quietrun) fprintf(stderr,"WARNING: disabling symbolic links due to relative paths\n");
	      nosymlinks=1;
	    }
	  sprintf(buffer,"find \"%s\" -type f %s -print", argv[i],(findopta==NULL)?"":findopta);
	  liste = popen (buffer, "r");
	  if(liste == 0)
	  {
	      perror(buffer);
	      return(-1);
	  }
	  filecount = readtree( &file_info, &frdinfosize, filecount, liste );
	  pclose(liste);
	}
    }
    if(interact)
    {
	/* worst case estimate is that all files are duplicate */
	dups_info = (dinfo*)calloc( (filecount>>1)+2, sizeof(dinfo) );
	if( dups_info == NULL )
	{
	    perror("alloc() does not provide info structure for duplicate files");
	    exit(-1);
	}
    }
    if(!quietrun) fprintf(stderr,"%d files to investigate\n\n",filecount);
    do
    {
	remember=totalinked-lastlinked;
        lastlinked=totalinked;
        qsort( file_info, filecount, sizeof(frdinfo), comparison_by_size );
#ifdef DEBUG
        if(!quietrun) 
	{
	    fprintf(stderr,"sorting loop ended with %d:%d\n",lastlinked,totalinked);
            print_infolist();
	}
#endif /* DEBUG */
    } while(lastlinked<totalinked && remember!=(totalinked-lastlinked) && ! testmode);
    if(!quietrun&&!interact) printstats();
    if(interact)  print_dupinfo();
    exit(exit_status);
}
